/*
 * Access a filesystem.
 *
 * Copyright (C) 2018, 2019, 2020 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA
 */

#include <l4/sys/types.h>

#include <sys/types.h>

#include "file_client.h"
#include "filesystem_client.h"
#include "mapped_file_client.h"
#include "opener_client.h"
#include "opener_context_client.h"
#include "pipe_client.h"
#include "pipe_opener_client.h"
#include "sync_client.h"

#include "fsdesc.h"
#include "fsdesc_client.h"
#include "fsops.h"

#include <ipc/mem_ipc.h>    /* ipc_dataspace functions */
#include <systypes/user.h>
#include <systypes/stat.h>

/* Filesystem configuration methods. */

long
fs_getuserfs(l4_cap_idx_t server, user_t user, l4_cap_idx_t *ref)
{
  return Filesystem_getuserfs((ref_Filesystem) server, user.uid, user.gid, user.umask, ref);
}



/* Filesystem context access methods. */

long
fs_context(file_descriptor_t *desc)
{
  long err = Opener_context((ref_Opener) desc->server, &desc->context);

  if (err)
    return err;

  /* Initialise the descriptor with the context as dataspace. */

  return fsdesc_init_buffer_from_dataspace(&desc->cobj, desc->context);
}

long
fs_context_getfs(file_descriptor_t *desc)
{
  return OpenerContext_getfs((ref_OpenerContext) desc->context, &desc->ref);
}

long
fs_context_mount(file_descriptor_t *desc)
{
  return OpenerContext_mount((ref_OpenerContext) desc->context, desc->ref);
}

long
fs_context_open(file_descriptor_t *desc, int flags)
{
  /* Use the context to obtain a file reference. */

  return OpenerContext_open((ref_OpenerContext) desc->context, flags,
                            &desc->obj.size, &desc->ref);
}



/* Filesystem object access methods. */

long
fs_begin(file_descriptor_t *desc)
{
  long err = Sync_begin((ref_Sync) desc->ref, &desc->obj.size);
  if (err)
    return err;

  /* Update state from the server. */

  fs_update_from_size(desc);

  return L4_EOK;
}

long
fs_begin_first(file_descriptor_t *desc)
{
  long err = Sync_begin_first((ref_Sync) desc->ref, &desc->obj.size);
  if (err)
    return err;

  /* Update state from the server. */

  fs_update_from_size(desc);

  return L4_EOK;
}

long
fs_end(file_descriptor_t *desc)
{
  return Sync_end((ref_Sync) desc->ref);
}

long
fs_flush(file_descriptor_t *desc)
{
  long err = File_flush((ref_File) desc->ref, desc->data_start, &desc->obj.size);
  if (err)
    return err;

  /* Update state from the server. */

  fs_update_from_size(desc);

  return L4_EOK;
}

long
fs_fstat(file_descriptor_t *desc, struct stat *statbuf)
{
  size_t size;
  l4_cap_idx_t statcap;
  sys_stat_t *buffer;
  unsigned long buffer_size;

  /* Request the metadata. */

  long err = File_fstat((ref_File) desc->ref, &size, &statcap);

  if (err)
    return err;

  if (size != sizeof(sys_stat_t))
    return -L4_EIO;

  /* Access the returned capability as a dataspace. */

  err = ipc_dataspace_size(statcap, &buffer_size);

  if (err)
    return err;

  err = ipc_attach_dataspace(statcap, buffer_size, (void **) &buffer);

  if (err)
    return err;

  /* Read the data from the message. */
  /* Convert the neutral representation to struct stat. */

  statbuf->st_dev = buffer->st_dev;
  statbuf->st_ino = buffer->st_ino;
  statbuf->st_mode = buffer->st_mode;
  statbuf->st_nlink = buffer->st_nlink;
  statbuf->st_uid = buffer->st_uid;
  statbuf->st_gid = buffer->st_gid;
  statbuf->st_rdev = buffer->st_rdev;
  statbuf->st_size = buffer->st_size;
  statbuf->st_atim.tv_sec = buffer->st_atim.tv_sec;
  statbuf->st_atim.tv_nsec = buffer->st_atim.tv_nsec;
  statbuf->st_mtim.tv_sec = buffer->st_mtim.tv_sec;
  statbuf->st_mtim.tv_nsec = buffer->st_mtim.tv_nsec;
  statbuf->st_ctim.tv_sec = buffer->st_ctim.tv_sec;
  statbuf->st_ctim.tv_nsec = buffer->st_ctim.tv_nsec;
  statbuf->st_blksize = buffer->st_blksize;

  /* Release the dataspace. */

  ipc_detach_dataspace(buffer);

  return L4_EOK;
}

long
fs_mmap(file_descriptor_t *desc, off_t position, size_t length)
{
  long err = MappedFile_mmap((ref_MappedFile) desc->ref, position, length,
                             &desc->obj.start_pos, &desc->obj.end_pos,
                             &desc->data_end);

  if (err)
    return err;

  /* Update the data start indicator. */
  /* Set the dataspace size. */

  desc->data_start = position - desc->obj.start_pos;
  desc->obj.buffer.size = desc->obj.end_pos - desc->obj.start_pos;

  return L4_EOK;
}

long
fs_yield(file_descriptor_t *desc)
{
  long err = Sync_yield((ref_Sync) desc->ref, &desc->obj.size);
  if (err)
    return err;

  /* Update state from the server. */

  fs_update_from_size(desc);

  return L4_EOK;
}



/* Pipe access methods. */

long
fs_pipe(l4_cap_idx_t opener, file_descriptor_t *reader, file_descriptor_t *writer)
{
  long err = PipeOpener_pipe((ref_PipeOpener) opener, &reader->ref, &writer->ref);
  if (err)
    return err;

  err = fs_current_region(reader);
  if (err)
    return err;

  return fs_current_region(writer);
}



/* Pipe access methods. */

long
fs_current_region(file_descriptor_t *desc)
{
  /* Reset the data extents in the region, updated by flushing. */

  fsdesc_reset_buffer(desc, 0);

  return Pipe_current_region((ref_Pipe) desc->ref, &desc->obj.buffer.size);
}

long
fs_next_region(file_descriptor_t *desc)
{
  /* Reset the data extents in the region, updated by flushing. */

  fsdesc_reset_buffer(desc, 0);

  return Pipe_next_region((ref_Pipe) desc->ref, &desc->obj.buffer.size, &desc->data_end);
}
